### Code Package for FlappyBird Q-Learning

---

#### 1. File structure

```shell
flappy-bird-qlearning
├── README.md
├── game
│   ├── LICENSE
│   ├── Pipfile
│   ├── Pipfile.lock
│   ├── README.md
│   ├── assets
│   ├── flappy.ico
│   ├── flappy.py
│   ├── screenshot1.png
│   └── setup.py
└── qlearning-bot
    ├── LICENSE
    ├── README.md
    ├── data
    ├── requirements.txt
    └── src
```



* `flappy-bird-qlearning/game` is the folder for the pure FlappyBird game 

* `flappy-bird-qlearning/qlearning-bot` is the folder for the RL-involved game
  * Only dependency of the project is `pygame`.




#### 2. How to install

Dependencies:

```
pygame==2.5.0
Python==3.11
```

You can install `pygame` by running:

```
pip install pygame==2.5.0
```



If no `pip` installed, please refer to [pip Installation](https://pip.pypa.io/en/stable/installation/). 



#### 4. How to run

To run the game only:

```
cd flappy-bird-qlearning/game
python flappy.py 
```



To run the Q-learning bot:

```
cd flappy-bird-qlearning/qlearning-bot
python src/flappy.py # the RL agent will play the game
```

- `flappy-bird-qlearning/qlearning-bot/src/flappy.py` - Run to see the actual visual gameplay.
- `flappy-bird-qlearning/qlearning-bot/src/learn.py` - Run for faster learning/training. This runs without any pygame visualization, so it's much faster.
  - The following command-line args are available:
    - `--verbose` to see `iteration | score` pair printed at each iteration. (Iteration = a bird playing from start until death)
    - `--iter` number of iterations to run.
- `flappy-bird-qlearning/qlearning-bot/src/initialize_qvalues.py` - Run if you want to reset the q-values, so you can observe how the bird learns to play over time.
- `flappy-bird-qlearning/qlearning-bot/src/bot.py` - This file contains the `Bot` class that applies the Q-Learning logic to the game.



#### 5. References

[1] http://sarvagyavaish.github.io/FlappyBirdRL/

[2] https://github.com/chapter09/flappybird-qlearning-bot

[3] http://artint.info/html/ArtInt_265.html









